/**
 ** MPC_plannerExecTaskCodels.cc
 **
 ** Codels called by execution task MPC_plannerExecTask
 **
 ** Author: Vivien Delsart
 ** Date: Tue Feb 22 16:18:09 CET 2011
 **
 **/

#include <portLib.h>

#include "server/MPC_plannerHeader.h"

#include <MPC/common.hpp>
#include <MPC/system.hpp>
#include <MPC/trajGen.hpp>
#include <MPC/planner.hpp>
#include <MPC/gnuplot.hpp>
using namespace mpc;

#include <iostream>
#include <cstring>
using namespace std;

// #define DEBUG 1

#define DFLT_PL_MODE SIMPLE_PL_MODE
#define DFLT_PL_OUTPUT PATH_PL_OUTPUT

System* sys;                         // the robotic system managed by MPC
SimpleSTrajPlanner* simplePlanner;   // the simple planner managed by MPC

const PSTraj* result_straj;          // the resulting state trajectory

POSTER_ID pomPosterId;               // POM poster ID (containing the localization of the robot)
POSTER_ID ctrlPosterId;              // GENPOS_CART_SPEED poster ID (containing the control of the robot)
POSTER_ID dtmPosterId;               // DTM poster ID (containing the DTM of the environment surrounding the robot)

/* Path recording variables */
float minDistBtwnPos;                // minimal distance between two positions of the recorded path
/* Traj recording variables */
float dt_trajRec;                    // minimal timebetween two state-times of the recorded trajectory

mpc::Timer timer_trajRec;            // timer used for the trajectory recording process

/*------------------------------------------------------------------------
 *  poster reading / writing tools
 */

/**
 *  \brief Convert a PL_PATH_XY poster into a DPMPath
 *  \param poster : the input poster
 *  \return a new DPMPath corresponding to the input poster
 *
 *  A new instance of DPMPath object is created
 */
DPMPath* PL_PATH_XYstruct_to_DPMPath(const PL_PATH_XY& poster)
{
  // create a new path
  DPMPath* newPath = new DPMPath();

  int N = poster.nbPos;

  for(int i=0; i<N; i++)
    {
      // get back the current configuration of the path
      GENPOS_CART_CONFIG pos = poster.tab[i];

      // create a MPC configuration from it
      DPMConfig cfg(pos.x,pos.y,pos.theta);

      // append it to the new path
      newPath->append(cfg);
    }

  // return the built path
  return newPath;
}

/**
 *  \brief Convert a MANA_STRAJ poster into a DD2STraj
 *  \param poster : the input poster
 *  \return a new DD2STraj corresponding to the input poster
 *
 *  A new instance of DD2STraj object is created
 */
DD2STraj* MANA_STRAJstruct_to_DD2STraj(const PL_MANA_STRAJ& poster)
{
  // create a new path
  DD2STraj* newTraj = new DD2STraj();

  int N = poster.nbPos;

  for(int i=0; i<N; i++)
    {
      // get back the current state-time of the trajectory
      PL_MANA_STATET statet = poster.tab[i];

      // create a MPC configuration from it
      DD2StateT s(statet.x,statet.y,statet.th,statet.v,statet.w,statet.t);

      // append it to the new path
      newTraj->append(s);
    }

  // return the built path
  return newTraj;
}

/*
 * \brief fill the current PL_PATH_XY structure from the mpc path
 * \param path_mpc : the input mpc path
 * \param path_gen : the resulting genom path
 */
void fill_PL_PATH_XY(const DPMPath& path_mpc, PL_PATH_XY& path_gen)
{
  int N = path_mpc.size();

//   cout << "==========================================================================" << endl;
//   path_mpc.display();
//   cout << "==========================================================================" << endl;

  for(int i=0; i<N; i++)
    {
      // get back the current configuration of the path
      const mpc::DPMConfig& cfg = path_mpc.at(i);

      // convert it into a GENPOS_CART_CONFIG
      path_gen.tab[i].x = cfg.x();
      path_gen.tab[i].y = cfg.y();
      path_gen.tab[i].theta = cfg.th();
    }
  path_gen.nbPos = N;

  return;
}

/*
 * \brief concatenate a mpc path to the current PL_PATH_XY structure
 * \param path_mpc : the input mpc path
 * \param path_gen : the resulting genom path
 */
void concat_PL_PATH_XY(const DPMPath& path_mpc, PL_PATH_XY& path_gen)
{
  int N = path_mpc.size();

  for(int i=0; i<N; i++)
    {
      if(path_gen.nbPos == PL_PLAN_MAXSIZE) { cerr << "MPC_planner::concat_PL_PATH_XY : Cannot concatenate the path, the target path is full." << endl; return; }

      // get back the current configuration of the path
      const mpc::DPMConfig& cfg = path_mpc.at(i);

      if(i != 0 || path_gen.nbPos == 0 || path_gen.tab[path_gen.nbPos-1].x != cfg.x() || path_gen.tab[path_gen.nbPos-1].y != cfg.y())
	{
	  // convert it into a GENPOS_CART_CONFIG
	  path_gen.tab[path_gen.nbPos].x = cfg.x();
	  path_gen.tab[path_gen.nbPos].y = cfg.y();
	  path_gen.tab[path_gen.nbPos].theta = cfg.th();
	  
	  path_gen.nbPos++;
	}
    }

  return;
}

/*
 *  \brief display the Mana traj poster structure
 */
void displayPL_PATH_XY(const PL_PATH_XY& path_gen)
{
  int N =  path_gen.nbPos;

  cout << "PATH POSTER STRUCTURE : " << endl;
  for(int i=0; i<N; i++)
    {
      cout << "(" << SDI_F->pathRef.tab[i].x << "," << SDI_F->pathRef.tab[i].y << "," << SDI_F->pathRef.tab[i].theta << ") ";
    }
  cout << endl;
}

// /*
//  * \brief fill the current PL_PATH_XY structure from the mpc path
//  * \param path_xy : the input mpc path
//  */
// void fill_MANA_STRAJ(const DD2STraj& dd2STraj)
// {
//   int N = dd2STraj.size();

//   for(int i=0; i<N; i++)
//     {
//       // get back the current configuration of the path
//       const mpc::DD2StateT& s = dd2STraj.at(i);

//       // convert it into a GENPOS_CART_CONFIG
//       SDI_F->mana_strajRef.tab[i].x = s.x();
//       SDI_F->mana_strajRef.tab[i].y = s.y();
//       SDI_F->mana_strajRef.tab[i].th = s.th();
//       SDI_F->mana_strajRef.tab[i].v = s.v();
//       SDI_F->mana_strajRef.tab[i].w = s.w();
//       SDI_F->mana_strajRef.tab[i].t = s.t();
//     }
//   SDI_F->mana_strajRef.nbPos = N;

//   return;
// }

/*
 * \brief fill the current PL_PATH_XY structure from the mpc path
 * \param path_xy : the input mpc path
 */
void fill_MANA_STRAJ(const DD2STraj& dd2STraj, PL_MANA_STRAJ& straj_gen)
{
  int N = dd2STraj.size();

  for(int i=0; i<N; i++)
    {
      // get back the current configuration of the path
      const mpc::DD2StateT& s = dd2STraj.at(i);

      // convert it into a GENPOS_CART_CONFIG
      straj_gen.tab[i].x = s.x();
      straj_gen.tab[i].y = s.y();
      straj_gen.tab[i].th = s.th();
      straj_gen.tab[i].v = s.v();
      straj_gen.tab[i].w = s.w();
      straj_gen.tab[i].t = s.t();
    }
  straj_gen.nbPos = N;

  return;
}


/*
 *  \brief display the Mana traj poster structure
 */
void display_MANA_TRAJ()
{
  int N =  SDI_F->mana_strajRef.nbPos;

  cout << "MANA TRAJ POSTER STRUCTURE : " << endl;
  for(int i=0; i<N; i++)
    {
      cout << "(" << SDI_F->mana_strajRef.tab[i].x << "," << SDI_F->mana_strajRef.tab[i].y << "," << SDI_F->mana_strajRef.tab[i].th 
	   << "," << SDI_F->mana_strajRef.tab[i].v << "," << SDI_F->mana_strajRef.tab[i].w << ") ";
    }
  cout << endl;
}


/*------------------------------------------------------------------------
 *
 * MPC_Pl_ExecTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_Pl_ExecTaskInit(int *report)
{
  return OK;
}

/*------------------------------------------------------------------------
 *
 * MPC_Pl_ExecTaskEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_Pl_ExecTaskEnd(void)
{
  return OK;
}

/*------------------------------------------------------------------------
 *
 * MPC_Pl_ExecTaskPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 * 
 * Returns:    OK or ERROR
 */

STATUS
MPC_Pl_ExecTaskPerm(int *report)
{
  return OK;
}
/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_POSTER_NOT_FOUND
 */

/* MPC_Pl_InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_InitMain(PL_INIT_I *Init_I, int *report)
{
  SDI_F->ctrl_poster_found = false;

  // check for the localization poster (POM)
  if (posterFind(Init_I->pomPoster, &pomPosterId) == ERROR) 
    {
      cerr << "MPC_planner_ExecTask: cannot find pom poster" << endl;
      *report = S_MPC_planner_POSTER_NOT_FOUND;
      return ETHER;
  }
#if DEBUG
  else cerr << "MPC_planner_ExecTask: pom poster found" << endl;
#endif

  // check for the Data Elevation Map poster (DTM)
  if (posterFind(Init_I->dtmPoster, &dtmPosterId) == ERROR) 
    {
      cerr << "MPC_planner_ExecTask: cannot find dtm poster" << endl;
      *report = S_MPC_planner_POSTER_NOT_FOUND;
      return ETHER;
  }
#if DEBUG
  else cerr << "MPC_planner_ExecTask: dtm poster found" << endl;
#endif

  // check for the control poster (GENPOS_CART_SPEED)
  if (posterFind(Init_I->ctrlPoster, &ctrlPosterId) == ERROR) 
    {
      cerr << "MPC_planner_ExecTask: cannot find ctrl poster" << endl;
  }
  else 
    {
#if DEBUG
      cerr << "MPC_planner_ExecTask: ctrl poster found" << endl;
#endif
      SDI_F->ctrl_poster_found = true;
    }

  /* Planning mode and output */
  SDI_F->pl_mode = DFLT_PL_MODE;
  SDI_F->pl_output = DFLT_PL_OUTPUT;

  /* Robotic system type */
  SDI_F->system_type = MPC_PLANNER_UNDEFINED_SYS;
  
  /* Path and trajectory structures */
  SDI_F->pathRec.nbPos = 0;
//   SDI_F->pathRec.corridorWidth = 0;
  SDI_F->mana_strajRec.nbPos = 0;
  SDI_F->pathRef.nbPos = 0;
//   SDI_F->pathRef.corridorWidth = 0;
  SDI_F->mana_strajRef.nbPos = 0;

  /* Init and goal locations */
  SDI_F->init_posXYTh.x = 0;
  SDI_F->init_posXYTh.y = 0;
  SDI_F->init_posXYTh.theta = 0;
  SDI_F->goal_posXYTh.x = 0;
  SDI_F->goal_posXYTh.y = 0;
  SDI_F->goal_posXYTh.theta = 0;

  result_straj = NULL;

  /* Path recording variables */
  minDistBtwnPos = 0.2;

  /* Trajectory recording variables */
  dt_trajRec = 0.5;
  
  return ETHER;
}


/*------------------------------------------------------------------------
 * PlanTo
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_NO_PATH
 *              S_MPC_planner_POM_POSTER_READ
 *              S_MPC_planner_DTM_POSTER_READ
 *              S_MPC_planner_UNDEFINED_SYS
 *              S_MPC_planner_GOAL_POS_OBS
 */

/* PlanToStart  -  codel START of PlanTo
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanToStart(PL_PLANTO_I *PlanTo_I, int *report)
{
  // check for the planning mode
  if(SDI_F->pl_mode == UNDEFINED_PL_MODE)
    {
      cerr << "MPC_planner: Undefined planning mode" << endl;
      return ETHER;
    }

  // check for the planning mode
  if(SDI_F->pl_output == UNDEFINED_PL_OUTPUT)
    {
      cerr << "MPC_planner: Undefined planning output" << endl;
      return ETHER;
    }

  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_plannerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
	{
	  cerr << "MPC_planner: Unable to read the POM poster" << endl;
	  return ETHER;
	}

  // set the init and goal positions
  SDI_F->init_posXYTh.x = pos.mainToOrigin.euler.x;
  SDI_F->init_posXYTh.y = pos.mainToOrigin.euler.y;
  SDI_F->init_posXYTh.theta = pos.mainToOrigin.euler.yaw;

  SDI_F->goal_posXYTh.x = PlanTo_I->goalX;
  SDI_F->goal_posXYTh.y = PlanTo_I->goalY;
  SDI_F->goal_posXYTh.theta = 0.0;

  // read informations about the robot's environment (DTM poster)
  // ...

  return EXEC;
}

/* PlanToMain  -  codel EXEC of PlanTo
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanToMain(PL_PLANTO_I *PlanTo_I, int *report)
{
  // create states from input init and goal positions
  ManaState initS(SDI_F->init_posXYTh.x, SDI_F->init_posXYTh.y, SDI_F->init_posXYTh.theta, 0.0, 0.0);
  ManaState goalS(SDI_F->goal_posXYTh.x, SDI_F->goal_posXYTh.y, SDI_F->goal_posXYTh.theta, 0.0, 0.0);
  
  // delete the old resulting state trajectory
  if(result_straj != NULL) delete result_straj;
  result_straj = NULL;

  // plan a state trajectory between both init and goal states
  int res = -1;
  switch(SDI_F->pl_mode)
    {
    case SIMPLE_PL_MODE :
      {
	res = ((ManaSimpleSTrajPlanner*)simplePlanner)->computeTraj(initS,goalS);
	
	if(res == 0) 
	  {
	    result_straj = ((ManaSimpleSTrajPlanner*)simplePlanner)->getPSTraj();
	    return END;
	  }
	else return FAIL;
      }
      
    default : 
      {
	cerr << "MPC_planner : Default planning mode. This case should not appear..." << endl;
	return FAIL;
      }
    }
}

/* PlanToEnd  -  codel END of PlanTo
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanToEnd(PL_PLANTO_I *PlanTo_I, int *report)
{
  cout << "MPC_planner : the planning process succeed : A path has been found" << endl;

  // fill the path poster
  if(result_straj != NULL) 
    {
      
      switch(SDI_F->pl_output)
	{
	case PATH_PL_OUTPUT : 
	  {
// 	    cout << "SAMPLING THE PATH..." << endl;

	    ManaSTraj* sampledSTraj = ((ManaSTraj*) (result_straj->sample(PL_PLAN_MAXSIZE)));
	    int N = sampledSTraj->size();

	    for(int i=0; i<N; i++)
	      {
		ManaStateT curST = sampledSTraj->at(i);
		SDI_F->pathRef.tab[i].x = curST.x();
		SDI_F->pathRef.tab[i].y = curST.y();
		SDI_F->pathRef.tab[i].theta = curST.th();

 		cout << "(" << SDI_F->pathRef.tab[i].x << "," << SDI_F->pathRef.tab[i].y << "," << SDI_F->pathRef.tab[i].theta << ") " << endl;
	      }
	    SDI_F->pathRef.nbPos = N;

//  	    cout << "SDI_F->pathRef.nbPos : " << SDI_F->pathRef.nbPos << endl;

	    delete sampledSTraj;
	    break;
	  }
	  
	  
	case STRAJ_PL_OUTPUT : 
	  {
//  	    cout << "SAMPLING THE TRAJECTORY..." << endl;

	    ManaSTraj* sampledSTraj = ((ManaSTraj*) (result_straj->sample(PL_PLAN_MAXSIZE)));
	    int N = sampledSTraj->size();

	    for(int i=0; i<N; i++)
	      {
		ManaStateT curST = sampledSTraj->at(i);
		SDI_F->mana_strajRef.tab[i].x = curST.x();
		SDI_F->mana_strajRef.tab[i].y = curST.y();
		SDI_F->mana_strajRef.tab[i].th = curST.th();
		SDI_F->mana_strajRef.tab[i].v = curST.v();
		SDI_F->mana_strajRef.tab[i].w = curST.w();
		SDI_F->mana_strajRef.tab[i].t = curST.t();

		SDI_F->pathRef.tab[i].x = curST.x();
		SDI_F->pathRef.tab[i].y = curST.y();
		SDI_F->pathRef.tab[i].theta = curST.th();
	      }
	    SDI_F->mana_strajRef.nbPos = N;
	    SDI_F->pathRef.nbPos = N;
//  	    cout << "SDI_F->mana_strajRef.nbPos : " << SDI_F->mana_strajRef.nbPos << endl;
	    
	    delete sampledSTraj;
	    break;
	  }
	  
	default : 
	  {
	    cerr << "MPC_planner : Default planning output. This case should not appear..." << endl;
	    return FAIL;
	  }
	}
    }

  return ETHER;
}

/* PlanToFail  -  codel FAIL of PlanTo
   Returns:  ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanToFail(PL_PLANTO_I *PlanTo_I, int *report)
{
  cout << "MPC_planner : the planning process failed : No path has been found" << endl;
  return ETHER;
}


/*------------------------------------------------------------------------
 * PlanSimplePath
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_NO_PATH
 *              S_MPC_planner_BLOCKED
 *              S_MPC_planner_POSTER_READ
 */

/* MPC_Pl_PlanSimplePathStart  -  codel START of PlanSimplePath
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimplePathStart(PL_PLANSIMPLEPATH_I *PlanSimplePath_I, int *report)
{
  // check if the robotic system has been defined
  if(SDI_F->system_type == MPC_PLANNER_UNDEFINED_SYS)
    {
      cerr << "MPC_planner_ExecTask: Cannot start the planning process : Undefined robotic system." << endl;      
      *report = S_MPC_planner_UNDEFINED_SYS;
      return FAIL;
    }

  // check if the input DEM is available
  // ...

  // set the init and goal positions
  SDI_F->init_posXYTh.x = PlanSimplePath_I->initPos.x;
  SDI_F->init_posXYTh.y = PlanSimplePath_I->initPos.y;
  SDI_F->init_posXYTh.theta = PlanSimplePath_I->initPos.theta;

  SDI_F->goal_posXYTh.x = PlanSimplePath_I->goalPos.x;
  SDI_F->goal_posXYTh.y = PlanSimplePath_I->goalPos.y;
  SDI_F->goal_posXYTh.theta = PlanSimplePath_I->goalPos.theta;

  return EXEC;
}

/* MPC_Pl_PlanSimplePathMain  -  codel EXEC of PlanSimplePath
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimplePathMain(PL_PLANSIMPLEPATH_I *PlanSimplePath_I, int *report)
{
  // read the input DEM
  // ...

  // create states from input init and goal positions
  ManaState initS(SDI_F->init_posXYTh.x, SDI_F->init_posXYTh.y, SDI_F->init_posXYTh.theta, 0.0, 0.0);
  ManaState goalS(SDI_F->goal_posXYTh.x, SDI_F->goal_posXYTh.y, SDI_F->goal_posXYTh.theta, 0.0, 0.0);
  
  // plan a state trajectory between both init and goal states
  int res = ((ManaSimpleSTrajPlanner*)simplePlanner)->computeTraj(initS,goalS);

  if(res == 0)
    {
      // case where the planning process succeed
      return END;
    }
  else 
    {
      // case where the planning process failed
      return FAIL;
    }
}

/* MPC_Pl_PlanSimplePathEnd  -  codel END of PlanSimplePath
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimplePathEnd(PL_PLANSIMPLEPATH_I *PlanSimplePath_I, int *report)
{
  cout << "MPC_planner : the planning process succeed : A path has been found" << endl;

  // fill the path poster
  const ManaSTraj* straj = ((ManaSimpleSTrajPlanner*)simplePlanner)->getSTraj(PL_PLAN_MAXSIZE);

  if(straj != NULL) 
    {
      int N = straj->size();
      for(int i=0; i<N; i++)
	{
	  ManaStateT curST = straj->at(i);
	  SDI_F->pathRef.tab[i].x = curST.x();
	  SDI_F->pathRef.tab[i].y = curST.y();
	  SDI_F->pathRef.tab[i].theta = curST.th();
	}
      SDI_F->pathRef.nbPos = N;
    }

  return ETHER;
}

/* MPC_Pl_PlanSimplePathFail  -  codel FAIL of PlanSimplePath
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimplePathFail(PL_PLANSIMPLEPATH_I *PlanSimplePath_I, int *report)
{
  cout << "MPC_planner : the planning process failed : No path has been found" << endl;

  return ETHER;
}


/*------------------------------------------------------------------------
 * PlanSimpleSTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_NO_PATH
 *              S_MPC_planner_BLOCKED
 *              S_MPC_planner_POSTER_READ
 */

/* MPC_Pl_PlanSimpleSTrajStart  -  codel START of PlanSimpleSTraj
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimpleSTrajStart(PL_PLANSIMPLESTRAJ_I *PlanSimpleSTraj_I, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* MPC_Pl_PlanSimpleSTrajMain  -  codel EXEC of PlanSimpleSTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimpleSTrajMain(PL_PLANSIMPLESTRAJ_I *PlanSimpleSTraj_I, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* MPC_Pl_PlanSimpleSTrajEnd  -  codel END of PlanSimpleSTraj
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimpleSTrajEnd(PL_PLANSIMPLESTRAJ_I *PlanSimpleSTraj_I, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}

/* MPC_Pl_PlanSimpleSTrajFail  -  codel FAIL of PlanSimpleSTraj
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_PlanSimpleSTrajFail(PL_PLANSIMPLESTRAJ_I *PlanSimpleSTraj_I, int *report)
{
  /* ... add your code here ... */
  return ETHER;
}


/*------------------------------------------------------------------------
 * SetManaSys
 *
 * Description: 
 *
 * Reports:      OK
 */

/* MPC_Pl_SetManaSysExec  -  codel EXEC of SetManaSys
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
MPC_Pl_SetManaSysExec(PL_SETMANASYS_I *SetManaSys_I, int *report)
{
  SDI_F->system_type = MPC_PLANNER_MANA_SYS;

#if DEBUG
  cerr << "MPC_planner_ExecTask: Mana system configured" << endl;
#endif

//   // fill the Mana system constraints structure
//   SDI_F->mana_sys.length = SetManaSys_I->body_length;
//   SDI_F->mana_sys.width = SetManaSys_I->body_width;
//   SDI_F->mana_sys.vMax = SetManaSys_I->vmax;
//   SDI_F->mana_sys.wMax = SetManaSys_I->wmax;
//   SDI_F->mana_sys.aMax = SetManaSys_I->accel;
//   SDI_F->mana_sys.uMax = SetManaSys_I->wmax;

  // Reset the local system
  if(sys != NULL) delete sys;
  sys = new ManaSys(SetManaSys_I->vmax,
		    SetManaSys_I->accel,
		    SetManaSys_I->wmax,
		    SetManaSys_I->wmax,
		    SetManaSys_I->body_length,
		    SetManaSys_I->body_width,
		    false);

  // Reset the controller with a simple controller for a Mana robotic system
  if(simplePlanner != NULL) delete simplePlanner;
#if DEBUG
  simplePlanner = new ManaSimpleSTrajPlanner((ManaSys*)sys, 0.8, 0.8, 1);
#else
  simplePlanner = new ManaSimpleSTrajPlanner((ManaSys*)sys, 0.8, 0.8, 0);
#endif

  return ETHER;
}


/*------------------------------------------------------------------------
 * StartPathRecording
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_REFPATH_POSTER_READ
 *              S_MPC_planner_POM_POSTER_READ
 */

/* StartPathRecordingStart  -  codel START of StartPathRecording
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartPathRecordingStart(PATHRECORDING_I *PathRecording_I, int *report)
{
  // reinit the path table size if needed
  SDI_F->pathRec.nbPos = 0;

  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_plannerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
    {
      cerr << "MPC_planner: Unable to read the POM poster" << endl;
      return FAIL;
    }
  
  // init the minimal distance between two positions of the recorded path
  minDistBtwnPos = PathRecording_I->minDistBtwnPos;

  // set the initial position of the robot path
  SDI_F->pathRec.tab[0].x = pos.mainToOrigin.euler.x;
  SDI_F->pathRec.tab[0].y = pos.mainToOrigin.euler.y;
  SDI_F->pathRec.tab[0].theta = pos.mainToOrigin.euler.yaw;
  SDI_F->pathRec.nbPos++;

#if DEBUG
  cout << "PATH RECORDING START"  << endl;
#endif

  /* ... add your code here ... */
  return EXEC;
}

/* StartPathRecordingMain  -  codel EXEC of StartPathRecording
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartPathRecordingMain(PATHRECORDING_I *PathRecording_I, int *report)
{
  // check if the maximal size of the sampled path tab has not been reached
  if(SDI_F->pathRec.nbPos >= PL_PLAN_MAXSIZE)
    {
      cerr << "MPC_planner: The maximal path tab size has been reached. Unable to continue recording positions." << endl;
      return END;
    }

  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_plannerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
    {
      cerr << "MPC_planner: Unable to read the POM poster" << endl;
      return FAIL;
    }

  float x = pos.mainToOrigin.euler.x;
  float y = pos.mainToOrigin.euler.y;
  float th = pos.mainToOrigin.euler.yaw;

  // compute the distance from the current position to the previous position
  double dx = x - SDI_F->pathRec.tab[SDI_F->pathRec.nbPos-1].x;
  double dy = y - SDI_F->pathRec.tab[SDI_F->pathRec.nbPos-1].y;
  double sqrtDist = dx*dx + dy*dy;

  // check if the distance is great enough to record the current position
  if(sqrtDist > minDistBtwnPos*minDistBtwnPos)
    {
      SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].x = x;
      SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].y = y;
      SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].theta = th;

//       cout << "PATH RECORDING EXEC : new position in tab (" << SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].x << "," << SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].y << "," << SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].theta << ")" << endl;

      SDI_F->pathRec.nbPos++;

#if DEBUG
      cout << "PATH RECORDING EXEC : new position (" << x << "," << y << "," << th << ") added (path size : " << SDI_F->pathRec.nbPos << ")"  << endl;
#endif

//       cout << "Current complete path SDI_F->pathRec.tab : " << endl;
//       for(int i=0; i<SDI_F->pathRec.nbPos; i++)
// 	{
// 	  cout << "(" << SDI_F->pathRec.tab[i].x << "," << SDI_F->pathRec.tab[i].y << "," << SDI_F->pathRec.tab[i].theta << ") -> ";
// 	}
//       cout << endl;

    }
#if DEBUG
  else cout << "PATH RECORDING EXEC : the current position (" << x << "," << y << "," << th << ") is too  close"  << endl;
#endif

  return EXEC;
}

/* StartPathRecordingEnd  -  codel END of StartPathRecording
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartPathRecordingEnd(PATHRECORDING_I *PathRecording_I, int *report)
{
  // building the ackerman path profile?
  // ...

#if DEBUG
  cout << "PATH RECORDING END... path size : " << SDI_F->pathRec.nbPos << endl;
#endif

  return ETHER;
}

/* StartPathRecordingFail  -  codel FAIL of StartPathRecording
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
StartPathRecordingFail(PATHRECORDING_I *PathRecording_I, int *report)
{
  // reinit the path table size if needed
  SDI_F->pathRec.nbPos = 0;

#if DEBUG
  cout << "PATH RECORDING FAIL"  << endl;
#endif

  return ETHER;
}

/* StartPathRecordingEnd  -  codel INTER of StartPathRecording
   see above (request StartPathRecording) */


/*------------------------------------------------------------------------
 * StartTrajRecording
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_CTRL_POSTER_READ
 *              S_MPC_planner_POM_POSTER_READ
 */

/* StartTrajRecordingStart  -  codel START of StartTrajRecording
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartTrajRecordingStart(TRAJRECORDING_I *TrajRecording_I, int *report)
{
  // reinit the timer
  timer_trajRec.reinit();

  // reinit the path table size if needed
  SDI_F->pathRec.nbPos = 0;

  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_plannerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
    {
      cerr << "MPC_planner: Unable to read the POM poster" << endl;
      return FAIL;
    }

  if(!SDI_F->ctrl_poster_found)
    {
      cerr << "MPC_planner: The control poster (GENPOS_CART_SPEED) has not been found. Cannot start the trajectory recording." << endl;
      return FAIL;
    }

  // read the current applied control (GENPOS_CART_SPEED poster)
  GENPOS_CART_SPEED ctrl;
  if(MPC_plannerGENPOS_CART_SPEEDPosterRead (ctrlPosterId, &ctrl) == ERROR)
    {
      cerr << "MPC_planner: Unable to read the GENPOS_CART_SPEED poster" << endl;
      return FAIL;
    }

  // init the minimal time between two state-times of the recorded trajectory
  dt_trajRec = TrajRecording_I->dt;
  
  // set the initial position of the robot path
  SDI_F->pathRec.tab[0].x = pos.mainToOrigin.euler.x;
  SDI_F->pathRec.tab[0].y = pos.mainToOrigin.euler.y;
  SDI_F->pathRec.tab[0].theta = pos.mainToOrigin.euler.yaw;
  SDI_F->pathRec.nbPos++;

  // set the initial position of the robot path
  SDI_F->mana_strajRec.tab[0].x = pos.mainToOrigin.euler.x;
  SDI_F->mana_strajRec.tab[0].y = pos.mainToOrigin.euler.y;
  SDI_F->mana_strajRec.tab[0].th = pos.mainToOrigin.euler.yaw;
  SDI_F->mana_strajRec.tab[0].v = ctrl.v;
  SDI_F->mana_strajRec.tab[0].w = ctrl.w;
  SDI_F->mana_strajRec.tab[0].t = 0;
  SDI_F->mana_strajRec.nbPos++;

  // start the timer
  timer_trajRec.start();

  cout << "TRAJ RECORDING START"  << endl;

  return EXEC;
}

/* StartTrajRecordingMain  -  codel EXEC of StartTrajRecording
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartTrajRecordingMain(TRAJRECORDING_I *TrajRecording_I, int *report)
{
  // check if the maximal size of the sampled path tab has not been reached
  if(SDI_F->mana_strajRec.nbPos >= PL_PLAN_MAXSIZE)
    {
      cerr << "MPC_planner: The maximal straj tab size has been reached. Unable to continue recording state-times." << endl;
      return END;
    }

  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_plannerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
    {
      cerr << "MPC_planner: Unable to read the POM poster" << endl;
      return FAIL;
    }

  // read the current applied control (GENPOS_CART_SPEED poster)
  GENPOS_CART_SPEED ctrl;
  if(MPC_plannerGENPOS_CART_SPEEDPosterRead (ctrlPosterId, &ctrl) == ERROR)
    {
      cerr << "MPC_planner: Unable to read the GENPOS_CART_SPEED poster" << endl;
      return FAIL;
    }

  float x = pos.mainToOrigin.euler.x;
  float y = pos.mainToOrigin.euler.y;
  float th = pos.mainToOrigin.euler.yaw;
  float v = ctrl.v;
  float w = ctrl.w;
  float t = timer_trajRec.totalElapsedTime();

  // get back the time of the last recorded state
  float lastT = SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos-1].t;

  // check if the distance is great enough to record the current position
  if(t - lastT > dt_trajRec)
    {
      SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].x = x;
      SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].y = y;
      SDI_F->pathRec.tab[SDI_F->pathRec.nbPos].theta = th;
      SDI_F->pathRec.nbPos++;

      SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos].x = pos.mainToOrigin.euler.x;
      SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos].y = pos.mainToOrigin.euler.y;
      SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos].th = pos.mainToOrigin.euler.yaw;
      SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos].v = ctrl.v;
      SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos].w = ctrl.w;
      SDI_F->mana_strajRec.tab[SDI_F->mana_strajRec.nbPos].t = timer_trajRec.totalElapsedTime();
      SDI_F->mana_strajRec.nbPos++;

#if DEBUG
      cout << "STRAJ RECORDING EXEC : new state-time (" << x << "," << y << "," << th << ")  ctrl : ("<< v << "," << w << ")  time : " << t << "    added (path size : " << SDI_F->mana_strajRec.nbPos << ")"  << endl;
#endif
    } 
#if DEBUG
  else cout << "STRAJ RECORDING EXEC : new state-time (" << x << "," << y << "," << th << ")  ctrl : ("<< v << "," << w << ")  time : " << t << "   is too close"  << endl;
#endif

  return EXEC;
}

/* StartTrajRecordingEnd  -  codel END of StartTrajRecording
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
StartTrajRecordingEnd(TRAJRECORDING_I *TrajRecording_I, int *report)
{
  // stop the timer
  timer_trajRec.stop();

#if DEBUG
  cout << "TRAJ RECORDING END... path size : " << SDI_F->pathRec.nbPos << "  traj size : " << SDI_F->mana_strajRec.nbPos << endl;
#endif

  return ETHER;
}

/* StartTrajRecordingFail  -  codel FAIL of StartTrajRecording
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
StartTrajRecordingFail(TRAJRECORDING_I *TrajRecording_I, int *report)
{
  // stop the timer
  timer_trajRec.stop();

  // reinit the path and traj table size if needed
  SDI_F->pathRec.nbPos = 0;
  SDI_F->mana_strajRec.nbPos = 0;

#if DEBUG
  cout << "PATH RECORDING FAIL"  << endl;
#endif

  return ETHER;
}

/* StartTrajRecordingEnd  -  codel INTER of StartTrajRecording
   see above (request StartTrajRecording) */


/*------------------------------------------------------------------------
 * LoadPath
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_FILE_NOT_FOUND
 */

/* LoadPathMain  -  codel EXEC of LoadPath
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LoadPathMain(LOADPATH_I *LoadPath_I, int *report)
{
  // reset the current reference path
  SDI_F->pathRef.nbPos = 0;

  // load a mpc path from the input filename
  DPMPath path_xy;
  bool resLoad = path_xy.load(LoadPath_I->filename);

  // check if the loading process succeeded
  if(!resLoad)
    {
      cerr << "MPC_planner::LoadPathMain : Unable to load the input path (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_NOT_FOUND;
      return END;
    }

  if(LoadPath_I->inverted)
    {
      // invert the current path
      DPMPath* inverted_path_xy = path_xy.invert();
      if(inverted_path_xy == NULL) { cerr << "MPC_planner::LoadPath : path inversion failed" << endl; return END; }

      // fill the current PL_PATH_XY structure from the mpc inverted path
      fill_PL_PATH_XY(*inverted_path_xy, SDI_F->pathRef);
      
      // delete the inverted path
      delete inverted_path_xy;
    }
  else
    {
      // fill the current PL_PATH_XY structure from the mpc path
      fill_PL_PATH_XY(path_xy, SDI_F->pathRef);
    }

  *report = OK;
  return END;
}

/*------------------------------------------------------------------------
 * LoadPathFromLoc
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_FILE_NOT_FOUND
 */

/* LoadPathFromLocStart  -  codel START of LoadPathFromLoc
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LoadPathFromLocStart(char path_filename[256], int *report)
{
  if(pomPosterId == NULL)
    {
      cerr << "MPCplanner::LoadPathFromLocStart : pom poster has not been found" << endl;
      return END;
    }

  return EXEC;
}

/* LoadPathFromLocMain  -  codel EXEC of LoadPathFromLoc
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LoadPathFromLocMain(char path_filename[256], int *report)
{
  // reset the current reference path
  SDI_F->pathRef.nbPos = 0;

  // load a mpc path from the input filename
  DPMPath path_mpc;
  bool resLoad = path_mpc.load(path_filename);

  // read the current location (POM poster)
  POM_POS pos;
  if(MPC_plannerPOM_POSPosterRead (pomPosterId, &pos) == ERROR)
	{
	  cerr << "MPC_planner: Unable to read the POM poster" << endl;
	  return ETHER;
	}

  // get back the first position of the path
  DPMConfig path_cfg0 = path_mpc.at(0);

  // move the origin of the path to the current location
  double dx = pos.mainToOrigin.euler.x - path_cfg0.x();
  double dy = pos.mainToOrigin.euler.y - path_cfg0.y();
  double dth = computeMinDiffAngle(pos.mainToOrigin.euler.yaw, path_cfg0.th());
  path_mpc.shiftXY(dx, dy);
  path_mpc.rotate(dth);

  // check if the loading process succeeded
  if(!resLoad)
    {
      cerr << "MPC_planner::LoadPathMain : Unable to load the input path (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_NOT_FOUND;
      return END;
    }

  // fill the current PL_PATH_XY structure from the mpc path
//   fill_PL_PATH_XY(path_xy);
  fill_PL_PATH_XY(path_mpc, SDI_F->pathRef);

  *report = OK;
  return END;
}


/*------------------------------------------------------------------------
 * SavePath
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_POSTER_NOT_FOUND
 *              S_MPC_planner_EMPTY_PATH
 *              S_MPC_planner_FILE_ERROR
 */

/* SavePathMain  -  codel EXEC of SavePath
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SavePathMain(char path_filename[256], int *report)
{
  // check if the current path is empty
  if(SDI_F->pathRec.nbPos == 0)
    {
      cerr << "MPC_planner::SavePathMain : The current path is empty. Cannot save it." << endl;
      *report = S_MPC_planner_EMPTY_PATH;
      return FAIL;
    }

  // convert the local structure into a mpc path
  DPMPath* path_xy = PL_PATH_XYstruct_to_DPMPath(SDI_F->pathRec);
  if(path_xy == NULL)
    {
      cerr << "MPC_planner::SavePathMain : The path conversion failed." << endl;
      *report = S_MPC_planner_EMPTY_PATH;
      return FAIL;
    }

  // save the path into the output file
  bool resSave = path_xy->save(path_filename);
  if(!resSave)
    {
      cerr << "MPC_planner::SavePathMain : Unable to save the current path (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_ERROR;
      return FAIL;
    }

  if(path_xy != NULL) delete path_xy;

  return END;
}


/*------------------------------------------------------------------------
 * ConcatPath
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_FILE_NOT_FOUND
 */

/* ConcatPathMain  -  codel EXEC of ConcatPath
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ConcatPathMain(LOADPATH_I *LoadPath_I, int *report)
{
  // load a mpc path from the input filename
  DPMPath path_xy;
  bool resLoad = path_xy.load(LoadPath_I->filename);

  // check if the loading process succeeded
  if(!resLoad)
    {
      cerr << "MPC_planner::LoadPathMain : Unable to load the input path (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_NOT_FOUND;
      return END;
    }

  if(LoadPath_I->inverted)
    {
      // invert the current path
      DPMPath* inverted_path_xy = path_xy.invert();
      if(inverted_path_xy == NULL) { cerr << "MPC_planner::LoadPath : path inversion failed" << endl; return END; }

      // fill the current PL_PATH_XY structure from the mpc inverted path
      concat_PL_PATH_XY(*inverted_path_xy, SDI_F->pathRef);
      
      // delete the inverted path
      delete inverted_path_xy;
    }
  else
    {
      // fill the current PL_PATH_XY structure from the mpc path
      concat_PL_PATH_XY(path_xy, SDI_F->pathRef);
    }

  *report = OK;
  return END;
}


/*------------------------------------------------------------------------
 * LoadTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_FILE_NOT_FOUND
 */

/* LoadTrajMain  -  codel EXEC of LoadTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LoadTrajMain(char path_filename[256], int *report)
{
  // reset the current reference path
  SDI_F->mana_strajRef.nbPos = 0;

  // load a mpc path from the input filename
  DD2STraj traj_mpc;
  bool resLoad = traj_mpc.load(path_filename);

  // check if the loading process succeeded
  if(!resLoad)
    {
      cerr << "MPC_planner::LoadTrajMain : Unable to load the input trajectory (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_NOT_FOUND;
      return END;
    }

  // fill the current MANA_STRAJ structure from the mpc traj
//   fill_MANA_STRAJ(traj_mpc);
  fill_MANA_STRAJ(traj_mpc, SDI_F->mana_strajRef);

  *report = OK;
  return END;
}

/*------------------------------------------------------------------------
 * SaveTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_POSTER_NOT_FOUND
 *              S_MPC_planner_EMPTY_TRAJ
 *              S_MPC_planner_FILE_ERROR
 */

/* SaveTrajMain  -  codel EXEC of SaveTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
SaveTrajMain(char path_filename[256], int *report)
{
  /* ... add your code here ... */
  return ETHER;
}


/*------------------------------------------------------------------------
 * PlanTrajPVD
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_EMPTY_PATH
 *              S_MPC_planner_PATH_TOO_SHORT
 */

/* PlanTrajPVDStart  -  codel START of PlanTrajPVD
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanTrajPVDStart(PVD_I *PlanTrajPVD_I, int *report)
{  
#if DEBUG
  cout << "PVD START" << endl;
#endif

  // check if the current path is not empty
  if(SDI_F->pathRef.nbPos == 0)
    {
      cerr << "PlanTrajPVDStart : The current path is empty. Cannot compute the trajectory." << endl;
      *report = S_MPC_planner_EMPTY_PATH;
      return FAIL;
    }

  // check if the input coefficient is valid
  if(PlanTrajPVD_I->coefVel <= 0 || PlanTrajPVD_I->coefVel > 1)
    {
      cerr << "PlanTrajPVDStart : Invalid input PVD coef : " << PlanTrajPVD_I->coefVel << " (must be in ]0;1]). Cannot compute the trajectory." << endl;
      return FAIL;
    }

  // check if the input robotic system has been defined
  if(sys == NULL)
    {
      cerr << "PlanTrajPVDStart : The robotic system has not been defined. Cannot compute the trajectory." << endl;
      return FAIL;
    }

  return EXEC;
}

/* PlanTrajPVDMain  -  codel EXEC of PlanTrajPVD
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanTrajPVDMain(PVD_I *PlanTrajPVD_I, int *report)
{
#if DEBUG
  cout << "PVD MAIN" << endl;
#endif

  SDI_F->mana_strajRef.nbPos = 0;

  // get back the computed path
  DPMPath* path = PL_PATH_XYstruct_to_DPMPath(SDI_F->pathRef);
  
  if(path != NULL)
    {    
      cout << "PlanTrajPVDMain : Path not null" << endl;

      // compute the velocity profile with PVD
      PVDTrajPlanner::setMSC(PlanTrajPVD_I->coefVel);
      PVDTrajPlanner::setMAC(PlanTrajPVD_I->coefAcc);

      DD2_PVDTrajPlanner planner((DiffDrive2Sys*)sys);
      int resPlan = planner.computeTraj(path);
      
      if(resPlan == TRAJ_CPTD)
	{
	  cout << "PlanTrajPVDMain : A trajectory has been computed!" << endl;
	  
	  // get back the planned trajectory
	  DD2STraj* straj = planner.getTraj();
	  
	  cout << "PlanTrajPVDMain PVD traj size : " << straj->size() << endl;

	  if(straj != NULL)
	    {
	      // fill the trajectory poster
// 	      fill_MANA_STRAJ(*straj);
	      fill_MANA_STRAJ(*straj, SDI_F->mana_strajRef);

	      cout << "PlanTrajPVDMain SDI_F traj size : " << SDI_F->mana_strajRef.nbPos << endl;
  
//  	      // display the path and velocity profiles
//  	      TrajPlot::plotDD2STraj(*straj, ((DiffDrive2Sys*)sys), true);
	      
 	      delete straj;
 	      delete path;

	      display_MANA_TRAJ();

	      cout << "PlanTrajPVDMain : A traj has been computed" << endl;
	      return END;
	    }
	  else 
	    {
	      delete path;
	      cout << "PlanTrajPVDMain : The traj is null" << endl;
	      return FAIL;
	    }
	}
      else 
	{
	  delete path;
	  cout << "PlanTrajPVDMain : The trajectory computation failed" << endl;
	  return FAIL;
	}
    }
  else 
    {
      cout << "PlanTrajPVDMain : No path computed" << endl;
      return FAIL;
    }
}

/* PlanTrajPVDEnd  -  codel END of PlanTrajPVD
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanTrajPVDEnd(PVD_I *PlanTrajPVD_I, int *report)
{
#if DEBUG
  cout << "PATH VEL DECOMPOSITION : END!" << endl;
#endif
  return ETHER;
}

/* PlanTrajPVDFail  -  codel FAIL of PlanTrajPVD
   Returns:  FAIL ZOMBIE */
ACTIVITY_EVENT
PlanTrajPVDFail(PVD_I *PlanTrajPVD_I, int *report)
{
  SDI_F->mana_strajRef.nbPos = 0;

#if DEBUG
  cout << "PATH VEL DECOMPOSITION : FAIL!" << endl;
#endif
  return ZOMBIE;
}

/* PlanTrajPVDInter  -  codel INTER of PlanTrajPVD
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
PlanTrajPVDInter(PVD_I *PlanTrajPVD_I, int *report)
{
  SDI_F->mana_strajRef.nbPos = 0;

#if DEBUG
  cout << "PATH VEL DECOMPOSITION : INTER!" << endl;
#endif
  return ETHER;
}


/*------------------------------------------------------------------------
 * RotateTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_POSTER_NOT_FOUND
 *              S_MPC_planner_EMPTY_TRAJ
 */

/* RotateTrajMain  -  codel EXEC of RotateTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
RotateTrajMain(float *rot_angle, int *report)
{
  // check if the reference path is not empty
  if(SDI_F->pathRef.nbPos == 0)
    {
      cerr << "RotateTrajMain : The reference path is empty!" << endl;
      return END;
    }

  // check if the reference trajectory is not empty
  if(SDI_F->mana_strajRef.nbPos == 0)
    {
      cerr << "RotateTrajMain : The reference trajectory is empty!" << endl;
      return END;
    }

  // get back a MPC path from the GEN path
  DPMPath* path_mpc = PL_PATH_XYstruct_to_DPMPath(SDI_F->pathRef);

  // check if the path is not null
  if(path_mpc != NULL)
    {
      // rotate the path
      float rot_angle_rad = (*rot_angle) / 180.0 * M_PI;
      path_mpc->rotate(rot_angle_rad);

      // fill the GEN path with it
      fill_PL_PATH_XY(*path_mpc, SDI_F->pathRef);

      // delete the MPC path
      delete path_mpc;
    }

  // get back a MPC trajectory from the GEN trajectory
  DD2STraj* traj_mpc = MANA_STRAJstruct_to_DD2STraj(SDI_F->mana_strajRef);

  // check if the trajectory is not null
  if(traj_mpc != NULL)
    {
      // rotate the trajectory
      float rot_angle_rad = (*rot_angle) / 180.0 * M_PI;
      traj_mpc->rotate(rot_angle_rad);

      // fill the GEN trajectory with it
      fill_MANA_STRAJ(*traj_mpc, SDI_F->mana_strajRef);

      // delete the MPC path
      delete traj_mpc;
    }

  return ETHER;
}

/*------------------------------------------------------------------------
 * ShiftLatTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_POSTER_NOT_FOUND
 *              S_MPC_planner_EMPTY_TRAJ
 */

/* ShiftLatTrajMain  -  codel EXEC of ShiftLatTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ShiftLatTrajMain(float *shift_dist, int *report)
{
  // check if the reference path is not empty
  if(SDI_F->pathRef.nbPos == 0)
    {
      cerr << "ShiftLatTrajMain : The reference path is empty!" << endl;
      return END;
    }

  // check if the reference trajectory is not empty
  if(SDI_F->mana_strajRef.nbPos == 0)
    {
      cerr << "ShiftLatTrajMain : The reference trajectory is empty!" << endl;
      return END;
    }

  // get back a MPC path from the GEN path
  DPMPath* path_mpc = PL_PATH_XYstruct_to_DPMPath(SDI_F->pathRef);

  // check if the path is not null
  if(path_mpc != NULL)
    {
      // shift the path
      path_mpc->shiftLat(*shift_dist);

      // fill the GEN path with it
      fill_PL_PATH_XY(*path_mpc, SDI_F->pathRef);

      // delete the MPC path
      delete path_mpc;
    }

  // get back a MPC trajectory from the GEN trajectory
  DD2STraj* traj_mpc = MANA_STRAJstruct_to_DD2STraj(SDI_F->mana_strajRef);

  // check if the trajectory is not null
  if(traj_mpc != NULL)
    {
      // rotate the trajectory
      traj_mpc->shiftLat(*shift_dist);

      // fill the GEN trajectory with it
      fill_MANA_STRAJ(*traj_mpc, SDI_F->mana_strajRef);

      // delete the MPC path
      delete traj_mpc;
    }

  return ETHER;
}


/*------------------------------------------------------------------------
 * ShiftXYTraj
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_EMPTY_TRAJ
 */

/* ShiftXYTrajMain  -  codel EXEC of ShiftXYTraj
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
ShiftXYTrajMain(SHIFT_XY_I *shift_xy, int *report)
{
  // check if the reference path is not empty
  if(SDI_F->pathRef.nbPos == 0)
    {
      cerr << "ShiftXYTrajMain : The reference path is empty!" << endl;
      return END;
    }

  // check if the reference trajectory is not empty
  if(SDI_F->mana_strajRef.nbPos == 0)
    {
      cerr << "ShiftXYTrajMain : The reference trajectory is empty!" << endl;
      return END;
    }

  // get back a MPC path from the GEN path
  DPMPath* path_mpc = PL_PATH_XYstruct_to_DPMPath(SDI_F->pathRef);

  // check if the path is not null
  if(path_mpc != NULL)
    {
      // shift the path
      path_mpc->shiftXY(shift_xy->dx, shift_xy->dy);

      // fill the GEN path with it
      fill_PL_PATH_XY(*path_mpc, SDI_F->pathRef);

      // delete the MPC path
      delete path_mpc;
    }

  // get back a MPC trajectory from the GEN trajectory
  DD2STraj* traj_mpc = MANA_STRAJstruct_to_DD2STraj(SDI_F->mana_strajRef);

  // check if the trajectory is not null
  if(traj_mpc != NULL)
    {
      // rotate the trajectory
      traj_mpc->shiftXY(shift_xy->dx, shift_xy->dy);

      // fill the GEN trajectory with it
      fill_MANA_STRAJ(*traj_mpc, SDI_F->mana_strajRef);

      // delete the MPC path
      delete traj_mpc;
    }

  return ETHER;
}


/*------------------------------------------------------------------------
 * LoadRobufastPath
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_FILE_NOT_FOUND
 */

/* LoadRobufatsPathMain  -  codel EXEC of LoadRobufastPath
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LoadRobufatsPathMain(FILENAME_I* Path_filename_I, int *report)
{
  // reset the current reference path
  SDI_F->pathRef.nbPos = 0;

  // load a mpc path from the input filename
  DPMPath path_xy;
  bool resLoad = path_xy.loadRobufastPath(Path_filename_I->filename);

  // check if the loading process succeeded
  if(!resLoad)
    {
      cerr << "MPC_planner::LoadPathMain : Unable to load the input path (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_NOT_FOUND;
      return END;
    }

  // fill the current PL_PATH_XY structure from the mpc path
  fill_PL_PATH_XY(path_xy, SDI_F->pathRef);

  *report = OK;
  return END;
}

/*------------------------------------------------------------------------
 * LoadRobufastPath2
 *
 * Description: 
 *
 * Reports:      OK
 *              S_MPC_planner_FILE_NOT_FOUND
 */

/* LoadRobufatsPathMain2  -  codel EXEC of LoadRobufastPath2
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
LoadRobufatsPathMain2(int *report)
{
  // reset the current reference path
  SDI_F->pathRef.nbPos = 0;

  // load a mpc path from the input filename
  DPMPath path_xy;
  bool resLoad = path_xy.loadRobufastPath("APPR2");

  // check if the loading process succeeded
  if(!resLoad)
    {
      cerr << "MPC_planner::LoadPathMain : Unable to load the input path (Wrong path filename?)" << endl;
      *report = S_MPC_planner_FILE_NOT_FOUND;
      return END;
    }
#if DEBUG
  cout << "resLoad robufast path : " << resLoad << endl;
#endif

  // fill the current PL_PATH_XY structure from the mpc path
  fill_PL_PATH_XY(path_xy, SDI_F->pathRef);

  displayPL_PATH_XY(SDI_F->pathRef);
  
  /* ... add your code here ... */
  return ETHER;
}
